import 'package:dynatrace_flutter_plugin/src/agent/util/logger.dart';
import 'package:http/http.dart';

import 'http_status_code_converter.dart';

/// Utils class which helps calculating values for web request monitoring
class HttpClientUtils {
  /// Calculating the length of the used [headers].
  static int getHeaderLength(Map<String, String> headers) {
    var stringBuffer = StringBuffer();

    headers.forEach((key, value) {
      stringBuffer.write("$key: ${value.toString()}\r\n");
    });

    return stringBuffer.length;
  }

  /// Calculation of the first line which appears in a request. This is mainly influenced
  /// by the [method], [path] and [query] of the URL.
  static int calcRequestFirstLineLength(
      String method, String path, String query) {
    var pathInfo = StringBuffer();
    pathInfo.write(path.isEmpty ? "/" : path);

    if (query.isNotEmpty) {
      pathInfo.write("?$query");
    }

    return ("$method ${pathInfo.toString()} HTTP/1.1").length;
  }

  /// Calculation the bytes of a request depending on the request [options].
  static int calcRequestSizeBytes(
      BaseRequest request, Map<String, String> additionalRequestHeaders) {
    // 4 Bytes coming from empty line after first line and headers
    int requestLength = 4;

    try {
      requestLength += calcRequestFirstLineLength(
          request.method, request.url.path, request.url.query);
      // Combining additional headers and request headers. Actual headers go second, so they override the default ones
      requestLength +=
          getHeaderLength({...additionalRequestHeaders, ...request.headers});

      if (request.contentLength != null && request.contentLength! > 0) {
        requestLength += request.contentLength!;

        if (!request.headers.containsKey("Content-Type")) {
          requestLength +=
              "Content-Type: application/x-www-form-urlencoded".length + 2;
        }

        if (!request.headers.containsKey("Content-Length")) {
          requestLength +=
              ("Content-Length: ${request.contentLength}").length + 2;
        }
      }

      if (!request.headers.containsKey("Host")) {
        requestLength += "Host: ${request.url.authority}".length + 2;
      }

      return requestLength;
    } catch (error) {
      Logger().d(
          "HttpClientUtils calcRequestSizeBytes(${request.url.toString()}): can't calculate request size.",
          logType: LogType.Warning);
      return -1;
    }
  }

  /// Calculate the response size depending on body of the [response]. As the content is stream,
  /// it can't be consumed twice, which means we need a content-length header to actually calculate
  /// the real response size bytes.
  static int calcResponseSizeBytes(StreamedResponse response) {
    try {
      // 4 Bytes coming from empty line after first line and headers
      int responseLength = 4;
      responseLength +=
          "HTTP/1.1 ${response.statusCode} ${HttpStatusCodeConverter.getStatusMessage(response.statusCode)}"
              .length;

      if (response.contentLength != null) {
        responseLength += response.contentLength!;
      }

      responseLength += getHeaderLength(response.headers);

      return responseLength;
    } catch (error) {
      var urlString = response.request != null
          ? response.request!.url.toString()
          : 'url not available';
      Logger().d(
          "HttpClientUtils calcResponseSizeBytes($urlString): can't calculate response size.",
          logType: LogType.Warning);
      return -1;
    }
  }
}
